# -*- coding: utf-8 -*-
import numpy as np
from scipy import constants
from scipy.optimize import minimize

from .theory import rotate_lines, theoretical_contour
from .deviation import radial_RMS

__all__ = ('young_laplace', 'fit')


class OptimizeResult(dict):
    """ Represents the optimization result.

    Notes
    -----
    This class has been copied from `scipy.optimize`.
    """
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())


def _drop_theo_points_outside_detection(R, Z, RZ_edges):
    """
    Remove theoretical points outside the range of detected ones.

    """
    R_edges, Z_edges = RZ_edges
    # Drop the theoretical points that go beyond the latest detected pixel
    # ie outside the image
    aa = np.where(Z > np.max(Z_edges))
    R = np.delete(R, aa)
    Z = np.delete(Z, aa)
    return R, Z


def young_laplace(surface_tension, angle, center_R, center_Z, radius, density_contrast,
                  calib, RZ_edges=None, *, gravity=None, num_points=1e3):
    """
    Returns the Young Laplace solution resized and oriented to the image.

    Parameters
    ----------
    surface_tension : scalar
        Surface tension
    angle : scalar
        Tilting angle.
    center_R : scalar
        Osculating circle center R-coordinate.
    center_Z : scalar
        Osculating circle center Z-coordinate.
    radius : scalar
        Osculating circle radius.
    density_contrast : scalar
        Difference of densities between the two fluids.
    calib : scalar
        Calibration in m per px.
    RZ_edges : tuple of array, optional
        (Radial, Vertical) coordinates of the edge.
        Used to crop the profile to the experimental profile.
        If None, the solution is not cropped.
    gravity : scalar, optional
        Gravitational acceleration. If None, `scipy.constants.g` is used.
    num_points : scalar, optional
        Number of points used in `theoretical_contour`

    Returns
    -------
    coordinates : tuple
        (R, Z)
    """
    if gravity is None:
        gravity = constants.g

    rho_g = density_contrast * gravity
    capillary_length = np.sqrt(surface_tension / rho_g)
    r0 = radius * calib
    bond_number = (r0 / capillary_length)**2

    # For the theoretical contour, the position (0, 0) corresponds to
    # the tip of the drop.
    R, Z = theoretical_contour(bond_number, num_points=num_points)

    # Rescale
    R = R * r0
    Z = Z * r0

    # Symmetrize the contour
    # Eliminate the first point as it is R=0, Z=0
    # To do not count it twice.
    # Then, flip the arrays to get points nicely ordered.
    R = np.concatenate((-R[1:][::-1], R))
    Z = np.concatenate((Z[1:][::-1], Z))

    # Rotate around the theoretical oscillating circle.
    base_center = np.array((0, r0))
    R, Z = rotate_lines(R, Z, base_center, angle)
    R = np.array(R)
    Z = np.array(Z)

    # Rescales contour to the image axes
    # The original drop apex is at (0, 0)
    # and is moved to the experimental apex
    R = R / calib + center_R
    Z = Z / calib + (center_Z - radius)

    if RZ_edges:
        R, Z = _drop_theo_points_outside_detection(R, Z, RZ_edges)
    return R, Z


def _compute_deviation(x0, *args, RMS=None):
    """
    Return the RMS for a profile given by set of parameters
    to the experimental profile.

    Parameters
    ----------
    x0 : tuple
        Initial parameters
    args : tuple
        Additional parameters

    The parameter order is:

    surface_tension : scalar
        Surface tension
    angle : scalar
        Tilting angle.
    center_R : scalar
        Osculating circle center R-coordinate.
    center_Z : scalar
        Osculating circle center Z-coordinate.
    radius : scalar
        Osculating circle radius.
    density_contrast : scalar
        Difference of densities between the two fluids.
    calib : scalar
        Calibration in m per px.
    RZ_edges : tuple of array, optional
        (Radial, Vertical) coordinates of the edge.
        Used to crop the profile to the experimental profile.
        If None, the solution is not cropped.

    `x0` and `args` must contain all these parameters.

    Returns
    -------
    RMS
    """
    RZ = young_laplace(*x0, *args)

    if RMS is None:
        RMS = radial_RMS

    RZ_edges = args[-1]
    return RMS(RZ, RZ_edges)


def fit(ini_variables,
        fixed_parameters,
        bounds,
        method='L-BFGS-B',
        options=None,
        ):
    """
    Minimize the root mean square distance.

    Parameters
    ----------
    ini_variables : tuple
        Initial guesses.
    fixed_parameters : tuple
        Parameters maintained constant during the minimization.
    bounds : tuple
        Upper and lower bounds for `ini_variables`.
        Must be of the same size.
    method : string, optional
        Method passed to `scipy.optimize.minimize`.
    options : dict, optional
        Options passed to `scipy.optimize.minimize`.

    Returns
    -------
    result : Instance of OptimizeResult class

    Notes
    -----
    The parameter order is:

    surface_tension : scalar
        Surface tension
    angle : scalar
        Tilting angle.
    center_R : scalar
        Osculating circle center R-coordinate.
    center_Z : scalar
        Osculating circle center Z-coordinate.
    radius : scalar
        Osculating circle radius.
    density_contrast : scalar
        Difference of densities between the two fluids.
    calib : scalar
        Calibration in m per px.
    RZ_edges : tuple of array, optional
        (Radial, Vertical) coordinates of the edge.
        Used to crop the profile to the experimental profile.
        If None, the solution is not cropped.

    The combination of `ini_variables` and `fixed_parameters`
    must contain all these parameters in this order.
    """
    if options is None:
        options = {'maxiter': 10,
                   'ftol': 1e-2,
                   'disp': False}
    res = minimize(_compute_deviation,
                   ini_variables,
                   args=fixed_parameters,
                   method=method,
                   bounds=bounds,
                   options=options)
    results = [x for x in res.x]
    results.extend(fixed_parameters[:-1])

    return OptimizeResult(surface_tension=results[0],
                          angle=results[1],
                          center_R=results[2],
                          center_Z=results[3],
                          radius=results[4],
                          RMS=res.fun)
