# -*- coding: utf-8 -*-
import numpy as np
import math


def print_parameters(title, params):
    """
    Print legibly the parameters.
    """
    if isinstance(params, dict):
        st = params['surface_tension']
        angle = params['angle']
        R = params['center_R']
        Z = params['center_Z']
        rad = params['radius']
        params.setdefault('RMS', None)
        RMS = params['RMS']
    else:
        st = params.surface_tension
        angle = params.angle
        R = params.center_R
        Z = params.center_Z
        rad = params.radius
        RMS = params.RMS


    print('')
    print('#' * 20)
    print(f'{title}')
    print('#' * 20)
    print(f'Surface tension: {st:.4} N/m')
    print(f'Tilt angle: {angle:.4}')
    print(f'Osculating circle center: {R:.2f}, {Z:.2f}')
    print(f'Osculating circle radius: {rad:.2f}')
    if RMS is not None:
        print(f'RMS: {RMS:.3}')
    print('')


def split_profile(R, Z):
    """
    Split a profile in two parts to get a single value for each Z.

    Parameters
    ----------
    R : array
        Radial coordinates.
    Z : array
        Vertical coordinates.
    """
    # Assumption on bubble upward orientation
    mask_left = R < R[Z.argmin()]
    R_left = R[mask_left]
    Z_left = Z[mask_left]
    R_right = R[~mask_left]
    Z_right = Z[~mask_left]
    return R_left, Z_left, R_right, Z_right


def rotate(x, y, cx, cy, angle):
    """
    Rotations of points.

    Parameters
    ----------
    x : array
        x-coordinates.
    y : array
        y-coordinates.
    cx : scalar
        x-center coordinate.
    cy : scalar
        y-center coordinate.
    angle : scalar
        Rotation angle in radian.

    Returns
    -------
    result : tuple
        (x_new, y_new)
    """
    vector = np.array((x - cx, y - cy))
    c, s = math.cos(angle), math.sin(angle)
    rotation_matrix = np.matrix([[c, s],
                                 [-s, c]])
    m = np.dot(vector.T, rotation_matrix)

    return m.T[0] + cx, m.T[1] + cy
