#!/usr/bin/env python3

import pytest

import os.path
import numpy as np

from drop.io import load_image
from drop.improcessing import fit_circle_tip, detect_edges, guess_angle
from drop.optimize import fit


def test_contour_ransac():
    image_path = os.path.join('data', 'uEye_Image_000827.png')
    zoom = ([100, 1312], [400, 1900])

    # Parameters
    calib = 0.00124 / 400  # m / px
    fluid_density = 1000  # kg/m3
    # Arbitrary first guess for gamma
    initial_surface_tension = 0.04  # N/m
    surface_tension_range = (0.02, 0.1)  # N/m

    image1 = load_image(image_path, region=zoom)

    edges, RZ_edges = detect_edges(image1)

    # Guess parameters
    center_Z, center_R, radius = fit_circle_tip(edges.shape,
                                                RZ_edges,
                                                method='ransac',
                                                debug=False)
    theta = guess_angle(RZ_edges, center_Z, center_R)

    ini_variables = (initial_surface_tension, )
    fixed_parameters = (theta, center_R, center_Z,
                        radius, fluid_density,
                        calib, RZ_edges)
    bounds = (surface_tension_range,)
    options = {'maxiter': 10,
               'ftol': 1e-2,
               'disp': False}
    # Here we use the simple model
    # Note that below the method method='SLSQP' can be used also.

    res = fit(ini_variables, fixed_parameters, bounds,
              method='L-BFGS-B', options=options, )

    ini_variables2 = (res.surface_tension,
                      res.angle,
                      res.center_R,
                      res.center_Z,
                      res.radius,
                      )
    bounds = ((res.surface_tension-1e-2, res.surface_tension+1e-2),
              (res.angle*0.7, res.angle*1.3),
              (res.center_R-5, res.center_R+5),
              (res.center_Z-5, res.center_Z+5),
              (res.radius-10, res.radius+10),
              )
    fixed_parameters = (fluid_density, calib, RZ_edges, )
    # Here, we use the full model
    options = {'maxiter': 100,
               'ftol': 1e-6,
               'disp': False}
    res = fit(ini_variables2, fixed_parameters, bounds,
              method='L-BFGS-B', options=options, )

    np.testing.assert_almost_equal(res.surface_tension, 0.07, decimal=3)
